/*
 * Infinity Portal by Craig A. Lindley
 *
 * This code uses the Tlc5940 Mux library available from
 * http://code.google.com/p/tlc5940arduino/ 
 * Copyright (c) 2009 by Alex Leone <acleone ~AT~ gmail.com>
 * with modifications by denos to make it work. See 
 * http://www.arduino.cc/cgi-bin/yabb2/YaBB.pl?num=1286580054
 *
 * Last Update: 03/03/2011
 *
 * The 32 RGB LEDs in the infinity portal are arranged as follows:
 *
 *                     23 22 21 20 19 18 17 16
 *                    __________________________
 *                    |                        |
 *                  24|                        |15
 *                  25|                        |14
 *                  26|                        |13
 *                  27|                        |12
 *                  28|                        |11
 *                  29|                        |10
 *                  30|                        |9
 *                  31|                        |8
 *                    |                        |
 *                    __________________________
 *                      0  1  2  3  4  5  6  7
 *
 * Where LEDs  0 -  7 are in row 0
 *       LEDs  8 - 15 are in row 1
 *       LEDs 16 - 23 are in row 2
 *       LEDs 24 - 31 are in row 3
 *
 * NOTE: A single RGB LED spans three consecutive channels of the TLC5940 with
 *       the RED portion (of the RGB LED) being the most significant channel.
 *
 * This code runs the infinity portal through a series of display patterns
 * that are selected at random. More patterns can be added at any time.
 */

#define LEDS_PER_ROW 8
#define XLAT  A4 
#define BLANK A5

#define  NUM_TLCS  2
#define  NUM_ROWS  4
#include "Tlc5940Mux.h"

volatile uint8_t isShifting;
uint8_t shiftRow, temp;

struct RGB {
  uint16_t red;
  uint16_t green;
  uint16_t blue;
};

// Palettes to be used by numerous patterns
struct RGB HSVPalette[32];
struct RGB FixPalette[7] = {
  0,   0,4095,
  0,4095,   0,
  0,4095,4095,
  4095,   0,   0,
  4095,   0,4095,
  4095,4095,   0,
  4095,4095,4095 
};

// Interrupt service routine
ISR(TIMER1_OVF_vect) {

  if (!isShifting) {
    isShifting = 1;

    // Disable all LEDs by pulling enable outputs high
    PORTC = 15;

    // Shift out the data for specified row
    TlcMux_shiftRow(shiftRow);

    // Set BLANK high
    digitalWrite(BLANK, HIGH);
    // 3 x 62.5ns = 187.5ns  (Blank needs to exceed 300ns to avoid shortening the GS cycle)
    __asm__("nop\n\t""nop\n\t""nop\n\t");

    // Set XLAT high
    digitalWrite(XLAT, HIGH);
    // XLAT for 62.5 ns
    __asm__("nop\n\t""nop\n\t");

    // Set XLAT low
    digitalWrite(XLAT, LOW);
    // Another 187.5 ns safely exceeding the minimum 300ns BLANK requirement    
    __asm__("nop\n\t""nop\n\t""nop\n\t");

    // Set BANK low
    digitalWrite(BLANK, LOW);
    // -- End THE IMPORTANT BIT -- //

    // Determine line of port C to pull low to enable row LEDs
    switch(shiftRow) {
    case 0: 
      temp = 14; 
      break;
    case 1: 
      temp = 13; 
      break;
    case 2: 
      temp = 11; 
      break;
    case 3: 
      temp = 7;  
      break;
    }
    // Turn the single bit row enable on 
    PORTC = temp;

    // Setup for next interrupt
    if (++shiftRow == NUM_ROWS) {
      shiftRow = 0;
    }
    isShifting = 0;
  }
}

int8_t adjLEDNum(int8_t ledNum) {

  if (ledNum < 0) {
    ledNum += 32;
  }  
  else if (ledNum > 31) {
    ledNum -= 32;
  }
  return ledNum;
}

// Get the row the specified LED resides in
uint8_t rowFromLED(uint8_t ledNum) {

  if ((ledNum < 0) || (ledNum > 31)) {
    Serial.println("Error");
    return -1;
  } 
  else if ((ledNum >= 0) && (ledNum <= 7)) {
    return 0;
  } 
  else if ((ledNum >= 8) && (ledNum <= 15)) {
    return 1;
  } 
  else if ((ledNum >= 16) && (ledNum <= 23)) {
    return 2;
  } 
  else {
    return 3;
  }    
}

// Return the blue channel number of the specified LED
uint8_t channelFromLED(uint8_t ledNum) {

  uint8_t row = rowFromLED(ledNum);
  uint8_t channelStart = row * LEDS_PER_ROW;
  uint8_t led = ledNum - channelStart;
  switch(led) {
  case 0: 
    return 0;
  case 1: 
    return 3;
  case 2: 
    return 6;
  case 3: 
    return 9;
  case 4: 
    return 16;
  case 5: 
    return 19;
  case 6: 
    return 22;
  case 7: 
    return 25;
  }
}

// Set the specifed LED to the specified Red, Green, Blue values
void setLEDRGB(uint8_t ledNum, uint16_t red, uint16_t green, uint16_t blue) {

  // Calculate row LED is in
  uint8_t row = rowFromLED(ledNum);

  // Calculate blue channel of LED
  uint8_t channel = channelFromLED(ledNum);

  // Set the three consecutive channels for this LED
  TlcMux_set(row, channel,   blue);
  TlcMux_set(row, channel+1, green);
  TlcMux_set(row, channel+2, red);
}

// Set the specifed LED to the value specified in the RGB structure
void setLEDRGB(uint8_t ledNum, struct RGB rgb) {
  setLEDRGB(ledNum, rgb.red, rgb.green, rgb.blue);
}

// Turn specified LED off (color is black)
void setLEDOff(uint8_t ledNum) {
  setLEDRGB(ledNum, 0, 0, 0);
}

// Color space conversion
// Input arguments
// hue in degrees (0 - 360.0)
// saturation in percent (0.0 - 1.0)
// value in percent (0.0 - 1.0)
// Output arguments
// red, green blue (0.0 - 1.0)
void _HSVtoRGB(double hue, double saturation, double value, double *red, double *green, double *blue) {

  int i;
  double f, p, q, t;
  if (saturation == 0) {
    // achromatic (grey)
    *red = *green = *blue = value;
    return;
  }
  hue /= 60;			// sector 0 to 5
  i = floor(hue);
  f = hue - i;			// factorial part of h
  p = value * (1 - saturation);
  q = value * (1 - saturation * f);
  t = value * (1 - saturation * (1 - f));
  switch(i) {
  case 0:
    *red = value;
    *green = t;
    *blue = p;
    break;
  case 1:
    *red = q;
    *green = value;
    *blue = p;
    break;
  case 2:
    *red = p;
    *green = value;
    *blue = t;
    break;
  case 3:
    *red = p;
    *green = q;
    *blue = value;
    break;
  case 4:
    *red = t;
    *green = p;
    *blue = value;
    break;
  default:		// case 5:
    *red = value;
    *green = p;
    *blue = q;
    break;
  }
}

// Convert HSV to RGB returning PWM values in range 0 .. 4095
void HSVtoRGB(float hue, float saturation, float value, uint16_t *red, uint16_t *green, uint16_t *blue) {
  double r, g, b;

  _HSVtoRGB(hue, saturation, value, &r, &g, &b);
  *red   = (uint16_t) (r * 4095);
  *green = (uint16_t) (g * 4095);
  *blue =  (uint16_t) (b * 4095);
}

// Create fully saturated color. Store in RGB structure
void createColor(float hue, float value, struct RGB *color) {
  double r, g, b;

  _HSVtoRGB(hue, 1.0, value, &r, &g, &b);
  (*color).red   = (uint16_t) (r * 4095);
  (*color).green = (uint16_t) (g * 4095);
  (*color).blue =  (uint16_t) (b * 4095);
}

// Create a palette of 32 HSV colors from across the hue spectrum
void createHSVPalette() {

  const float hueIncrement = 11.25;
  float hueAngle = 0.0;

  for (int i = 0; i < 32; i++) {
    createColor(hueAngle, 1.0, &HSVPalette[i]);
    hueAngle += hueIncrement;
  }
}

// Red then Green then Blue all the way around
void sequence1Pattern(uint8_t loops) {
  uint16_t red, green, blue;

  while (loops-- > 0) {

    TlcMux_clear();
    delay(2000);

    for (uint8_t color = 0; color < 3; color++) {
      red = green = blue = 0;
      switch(color) {
      case 0: 
        red   = 4095; 
        break;
      case 1: 
        green = 4095; 
        break;
      case 2: 
        blue  = 4095; 
        break;
      }

      for (uint8_t ledNum = 0; ledNum < 32; ledNum++) {
        setLEDRGB(ledNum, red, green, blue);
        delay(300);
      }
    }
  }
}

// Red then Green then Blue for each LED all the way around
void sequence2Pattern(uint8_t loops) {
  uint16_t red, green, blue;

  while (loops-- > 0) {

    TlcMux_clear();
    delay(2000);

    for (uint8_t ledNum = 0; ledNum < 32; ledNum++) {

      setLEDRGB(ledNum, FixPalette[3]);
      delay(250);
      setLEDRGB(ledNum, FixPalette[1]);
      delay(250);
      setLEDRGB(ledNum, FixPalette[0]);
      delay(250);
      setLEDOff(ledNum);
      delay(250);
    }
  }
}

// Pick random colors from FixPalette for each LED all the way around
void sequence3Pattern(uint8_t loops) {
  uint16_t red, green, blue, index;

  while (loops-- > 0) {

    TlcMux_clear();
    delay(2000);

    for (uint8_t ledNum = 0; ledNum < 32; ledNum++) {

      // Pick an index to decide LED color
      index = random(7);
      setLEDRGB(ledNum, FixPalette[index]);
      delay(300);
    }
  }
}

// Select random LEDs and give them random colors
void randomPattern(uint8_t loops) {
  uint16_t red, green, blue;
  uint8_t ledNum, delayAmount;

  TlcMux_clear();
  while (loops-- > 0) {
    ledNum = random(32);
    red = random(4096);
    green = random(4096);
    blue = random(4096);

    setLEDRGB(ledNum, red, green, blue);
    delay(150);
    setLEDOff(ledNum);
    delayAmount = random(100, 200);
    delay(delayAmount);
  }
}

// Two white LED chase each other
void chase1Pattern(uint8_t loops) {

  uint8_t led1Num, led2Num;

  led1Num =  0;
  led2Num = 16;

  TlcMux_clear();
  while (loops > 0) {
    setLEDRGB(led1Num, FixPalette[6]);
    setLEDRGB(led2Num, FixPalette[6]);
    delay(30);
    setLEDOff(led1Num);
    setLEDOff(led2Num);

    led1Num++;
    if (led1Num > 31) {
      led1Num = 0;
      --loops;
    }
    led2Num++;
    if (led2Num > 31) {
      led2Num = 0;
    }
  }
}

// Two multi-color LED chase each other
void chase2Pattern(uint8_t loops) {

  uint8_t led1Num, led2Num;
  float led1Angle, led2Angle;
  float degreeIncrement = 11.25;
  uint16_t red, green, blue;

  led1Num =  0;
  led2Num = 16;

  led1Angle = 0;
  led2Angle = led2Num * degreeIncrement;

  TlcMux_clear();
  while (loops > 0) {
    // Calculate led1 color
    HSVtoRGB(led1Angle, 1.0, 1.0, &red, &green, &blue);
    setLEDRGB(led1Num, red, green, blue);

    // Calculate led2 color
    HSVtoRGB(led2Angle, 1.0, 1.0, &red, &green, &blue);
    setLEDRGB(led2Num, red, green, blue);

    delay(60);
    setLEDOff(led1Num);
    setLEDOff(led2Num);

    led1Num++;
    if (led1Num > 31) {
      led1Num = 0;
      --loops;
    }
    led1Angle += degreeIncrement;
    if (led1Angle > 360.0) {
      led1Angle -= 360.0;
    }

    led2Num++;
    if (led2Num > 31) {
      led2Num = 0;
    }

    led2Angle += degreeIncrement;
    if (led2Angle > 360.0) {
      led2Angle -= 360.0;
    }
  }
}

// Four multi-color LED chase each other
void chase3Pattern(uint8_t loops) {

  int8_t led1Num, led2Num, led3Num, led4Num;
  float led1Angle, led2Angle, led3Angle, led4Angle;
  float degreeIncrement = 11.25;
  uint16_t red, green, blue;

  led1Num =  0;
  led2Num =  8;
  led3Num = 16;
  led4Num = 24;

  led1Angle = 0;
  led2Angle = led2Num * degreeIncrement;
  led3Angle = led3Num * degreeIncrement;
  led4Angle = led4Num * degreeIncrement;

  TlcMux_clear();
  while (loops > 0) {
    // Calculate led1 color
    HSVtoRGB(led1Angle, 1.0, 1.0, &red, &green, &blue);
    setLEDRGB(led1Num, red, green, blue);

    // Calculate led2 color
    HSVtoRGB(led2Angle, 1.0, 1.0, &red, &green, &blue);
    setLEDRGB(led2Num, red, green, blue);

    // Calculate led3 color
    HSVtoRGB(led3Angle, 1.0, 1.0, &red, &green, &blue);
    setLEDRGB(led3Num, red, green, blue);

    // Calculate led4 color
    HSVtoRGB(led4Angle, 1.0, 1.0, &red, &green, &blue);
    setLEDRGB(led4Num, red, green, blue);

    delay(60);

    setLEDOff(led1Num);
    setLEDOff(led2Num);
    setLEDOff(led3Num);
    setLEDOff(led4Num);

    led1Num--;
    if (led1Num < 0) {
      led1Num = 31;
      --loops;
    }
    led1Angle += degreeIncrement;
    if (led1Angle > 360.0) {
      led1Angle -= 360.0;
    }

    led2Num--;
    if (led2Num < 0) {
      led2Num = 31;
    }

    led2Angle += degreeIncrement;
    if (led2Angle > 360.0) {
      led2Angle -= 360.0;
    }

    led3Num--;
    if (led3Num < 0) {
      led3Num = 31;
    }

    led3Angle += degreeIncrement;
    if (led3Angle > 360.0) {
      led3Angle -= 360.0;
    }

    led4Num--;
    if (led4Num < 0) {
      led4Num = 31;
    }

    led4Angle += degreeIncrement;
    if (led4Angle > 360.0) {
      led4Angle -= 360.0;
    }
  }
}

// Flash all the LEDs at once in various colors
void flashPattern(uint8_t loops) {
  uint16_t red, green, blue, index;

  TlcMux_clear();

  while (loops-- > 0) {

    TlcMux_clear();
    delay(2000);

    // Pick an index to decide LED color
    index = random(7);

    // Set all of the LED to the same color
    for (int i = 0; i < 32; i++) {
      setLEDRGB(i, FixPalette[index]);
    }
    delay(150);
  }
}

// Rotate pattern
void rotate1Pattern(uint8_t loops) {

  uint16_t red, green, blue, index, colorIndex;

  TlcMux_clear();

  colorIndex = 0;

  while (loops-- > 0) {
    index = colorIndex;

    for (int i = 0; i < 32; i++) {

      setLEDRGB(i, HSVPalette[index]);

      if (++index > 31) {
        index = 0;
      }
    }
    delay(250);
    if (++colorIndex > 31) {
      colorIndex = 0;
    }
  }
}

// Rotate pattern
void rotate2Pattern(uint8_t loops) {

  TlcMux_clear();

  while (loops-- > 0) {

    for (int i = 0; i < 32; i++) {

      setLEDRGB(i, HSVPalette[i]);
      setLEDOff(adjLEDNum(i - 1));
      delay(250);
    }
  }
}

// Rotate pattern
void rotate3Pattern() {

  boolean isEven;

  TlcMux_clear();
  delay(1000);

  int colorIndex = 0;

  for (int x = 0; x < 28; x++) {
    isEven = ((x % 2) == 0);

    // Draw bottom LEDs
    setLEDRGB( 2, HSVPalette[colorIndex + 3]);
    setLEDRGB( 5, HSVPalette[colorIndex + 3]);
    if(isEven) {
      setLEDRGB( 1, HSVPalette[colorIndex + 3]);
      setLEDRGB( 6, HSVPalette[colorIndex + 3]);
    }  
    else  {
      setLEDRGB( 3, HSVPalette[colorIndex + 3]);
      setLEDRGB( 4, HSVPalette[colorIndex + 3]);
    }
    delay(200);

    // Draw right hand LEDs
    setLEDRGB(10, HSVPalette[colorIndex + 2]);
    setLEDRGB(13, HSVPalette[colorIndex + 2]);
    if(isEven) {
      setLEDRGB(11, HSVPalette[colorIndex + 2]);
      setLEDRGB(12, HSVPalette[colorIndex + 2]);
    }  
    else  {
      setLEDRGB( 9, HSVPalette[colorIndex + 2]);
      setLEDRGB(14, HSVPalette[colorIndex + 2]);
    }
    delay(200);

    // Draw top LEDs
    setLEDRGB(18, HSVPalette[colorIndex + 1]);
    setLEDRGB(21, HSVPalette[colorIndex + 1]);
    if(isEven) {
      setLEDRGB(17, HSVPalette[colorIndex + 1]);
      setLEDRGB(22, HSVPalette[colorIndex + 1]);
    }  
    else  {
      setLEDRGB(19, HSVPalette[colorIndex + 1]);
      setLEDRGB(20, HSVPalette[colorIndex + 1]);
    }
    delay(200);

    // Draw left hand LEDs
    setLEDRGB(26, HSVPalette[colorIndex]);
    setLEDRGB(29, HSVPalette[colorIndex]);
    if(isEven) {
      setLEDRGB(27, HSVPalette[colorIndex]);
      setLEDRGB(28, HSVPalette[colorIndex]);
    }  
    else  {
      setLEDRGB(25, HSVPalette[colorIndex]);
      setLEDRGB(30, HSVPalette[colorIndex]);
    }
    delay(200);

    colorIndex++;

    TlcMux_clear();
    delay(250);
  }
}

void rainPattern(uint8_t loops) {

  uint16_t red, green, blue, index, colorIndex;
  struct RGB colors[12];
  float degreeIncrement = 30.0;
  float hueAngle = 0;

  TlcMux_clear();

  // Calculate and store the colors
  for (int i = 0; i < 12; i++) {

    HSVtoRGB(hueAngle, 1.0, 1.0, &red, &green, &blue);
    colors[i].red = red;
    colors[i].green = green;
    colors[i].blue = blue;

    hueAngle += degreeIncrement;
  }
  int8_t led1Num, led2Num;

  while (loops-- > 0) {

    led1Num = 20;
    led2Num = 19;

    // Flow down
    for (int i = 0; i < 12; i++) {
      setLEDRGB(led1Num, colors[i]);
      setLEDRGB(led2Num, colors[i]);
      delay(250);
      setLEDOff(led1Num++);
      setLEDOff(led2Num--);
    }

    // Flow up
    led1Num = 3;
    led2Num = 4;

    for (int i = 0; i < 12; i++) {
      red = colors[i].red;
      green = colors[i].green;
      blue = colors[i].blue;

      setLEDRGB(led1Num, red, green, blue);
      setLEDRGB(led2Num, red, green, blue);
      delay(250);
      setLEDOff(led1Num--);
      setLEDOff(led2Num++);

      if (led1Num < 0) {
        led1Num = 31;
      }
    }
  }
}

// Alternating red/green colors
void redGreenPattern() {

  TlcMux_clear();
  delay(2000);

  for ( int i = 0; i < 32; i++) {
    if (i % 2 == 0) {
      setLEDRGB(i, FixPalette[3]);
    }  
    else  {
      setLEDRGB(i, FixPalette[1]);
    }
  }
  delay(5000);
}

// Alternating green/blue colors
void greenBluePattern() {

  TlcMux_clear();
  delay(2000);

  for ( int i = 0; i < 32; i++) {
    if (i % 2 == 0) {
      setLEDRGB(i, FixPalette[1]);
    }  
    else  {
      setLEDRGB(i, FixPalette[0]);
    }
  }
  delay(5000);
}

// Alternating red/blue colors
void redBluePattern() {

  TlcMux_clear();
  delay(2000);

  for ( int i = 0; i < 32; i++) {
    if (i % 2 == 0) {
      setLEDRGB(i, FixPalette[3]);
    }  
    else  {
      setLEDRGB(i, FixPalette[0]);
    }
  }
  delay(5000);
}

// Fill in pattern
void fillInPattern(uint8_t loops) {

  uint8_t ledNums[32] = {
    3,11,19,27,
    4,12,20,28,
    5,13,21,29,
    6,14,22,30,
    7,15,23,31,
    0,8,16,24,
    1,9,17,25,
    2,10,18,26
  };

  uint16_t index, ledIndex, colorIndex;

  TlcMux_clear();

  while (loops-- > 0) {

    TlcMux_clear();

    ledIndex = random(32);

    for (int i = 0; i < 32; i++) {
      colorIndex = random(7);
      index = ledNums[ledIndex];
      setLEDRGB(index, FixPalette[colorIndex]);
      delay(200);
      ledIndex++;
      if (ledIndex > 31) {
        ledIndex = 0;
      }
    }
  }
}

// Diverging lines pattern
void divergingLines1Pattern(uint8_t loops) {

  uint16_t led1Index, led2Index, colorIndex;

  while (loops-- > 0) {

    TlcMux_clear();
    delay(2000);

    led1Index = 3;
    led2Index = 4;

    for (int i = 0; i < 4; i++) {
      colorIndex = random(32);
      setLEDRGB(led1Index--, HSVPalette[colorIndex]);
      setLEDRGB(led2Index++, HSVPalette[colorIndex]);
      delay(300);
    }
  }
}

// Diverging lines pattern
void divergingLines2Pattern(uint8_t loops) {

  uint8_t led1Index, led2Index, led3Index, led4Index, led5Index, led6Index, led7Index, led8Index, colorIndex;

  while (loops-- > 0) {

    TlcMux_clear();
    delay(2000);

    led1Index = 3;
    led2Index = 4;
    led3Index = 11;
    led4Index = 12;
    led5Index = 19;
    led6Index = 20;
    led7Index = 27;
    led8Index = 28;

    for (int i = 0; i < 4; i++) {
      colorIndex = random(32);

      setLEDRGB(led1Index--, HSVPalette[colorIndex]);
      setLEDRGB(led2Index++, HSVPalette[colorIndex]);
      setLEDRGB(led3Index--, HSVPalette[colorIndex]);
      setLEDRGB(led4Index++, HSVPalette[colorIndex]);
      setLEDRGB(led5Index--, HSVPalette[colorIndex]);
      setLEDRGB(led6Index++, HSVPalette[colorIndex]);
      setLEDRGB(led7Index--, HSVPalette[colorIndex]);
      setLEDRGB(led8Index++, HSVPalette[colorIndex]);
      delay(300);
    }
  }
}

// Converging lines pattern
void convergingLines1Pattern(uint8_t loops) {

  uint16_t led1Index, led2Index, colorIndex;

  while (loops-- > 0) {

    TlcMux_clear();
    delay(2000);

    led1Index = 0;
    led2Index = 7;

    for (int i = 0; i < 4; i++) {
      colorIndex = random(32);
      setLEDRGB(led1Index++, HSVPalette[colorIndex]);
      setLEDRGB(led2Index--, HSVPalette[colorIndex]);
      delay(400);
    }
  }
}

// Converging lines pattern
void convergingLines2Pattern(uint8_t loops) {

  uint8_t led1Index, led2Index, led3Index, led4Index, led5Index, led6Index, led7Index, led8Index, colorIndex;

  while (loops-- > 0) {

    TlcMux_clear();
    delay(2000);

    led1Index = 0;
    led2Index = 7;
    led3Index = 8;
    led4Index = 15;
    led5Index = 16;
    led6Index = 23;
    led7Index = 24;
    led8Index = 31;

    for (int i = 0; i < 4; i++) {
      colorIndex = random(32);
      setLEDRGB(led1Index++, HSVPalette[colorIndex]);
      setLEDRGB(led2Index--, HSVPalette[colorIndex]);
      setLEDRGB(led3Index++, HSVPalette[colorIndex]);
      setLEDRGB(led4Index--, HSVPalette[colorIndex]);
      setLEDRGB(led5Index++, HSVPalette[colorIndex]);
      setLEDRGB(led6Index--, HSVPalette[colorIndex]);
      setLEDRGB(led7Index++, HSVPalette[colorIndex]);
      setLEDRGB(led8Index--, HSVPalette[colorIndex]);
      delay(400);
    }
  }
}

// Unidirectional pattern of LED with streamers
void streamer1Pattern() {

  struct RGB CV1, CV2, CV3, CV4;
  const float hueIncrement = 11.25;
  float hueAngle = 0.0;
  int8_t ledNum = 3;

  TlcMux_clear();

  // Do for 32 colors
  for (int c = 0; c < 32; c++) {

    // Calculate colors at the various percentages of brightness
    createColor(hueAngle, 1.0, &CV1);
    createColor(hueAngle, 0.4, &CV2);
    createColor(hueAngle, 0.2, &CV3);
    createColor(hueAngle, 0.1, &CV4);

    hueAngle += hueIncrement;

    // For each LED
    for (int i = 0; i < 32; i++) {
      setLEDRGB(adjLEDNum(ledNum),   CV1);
      setLEDRGB(adjLEDNum(ledNum-1), CV2);
      setLEDRGB(adjLEDNum(ledNum-2), CV3);
      setLEDRGB(adjLEDNum(ledNum-3), CV4);
      setLEDOff(adjLEDNum(ledNum-4));

      if (++ledNum > 31) {
        ledNum = 0;
      }
      delay(50);
    }
  }
}

// Unidirectional pattern of LED with streamers that speeds up
void streamer2Pattern() {

  struct RGB CV1, CV2, CV3, CV4;
  const float hueIncrement = 10.0;
  float hueAngle = 0.0;
  int16_t delayTimeMS, ledNum = 3;

  TlcMux_clear();

  delayTimeMS = 148;

  // Do for 32 colors
  for (int c = 0; c < 32; c++) {

    delayTimeMS -= 4;

    // Calculate colors at the various percentages of brightness
    createColor(hueAngle, 1.0, &CV1);
    createColor(hueAngle, 0.4, &CV2);
    createColor(hueAngle, 0.2, &CV3);
    createColor(hueAngle, 0.1, &CV4);

    hueAngle += hueIncrement;

    // For each LED
    for (int i = 0; i < 32; i++) {
      setLEDRGB(adjLEDNum(ledNum),   CV1);
      setLEDRGB(adjLEDNum(ledNum-1), CV2);
      setLEDRGB(adjLEDNum(ledNum-2), CV3);
      setLEDRGB(adjLEDNum(ledNum-3), CV4);
      setLEDOff(adjLEDNum(ledNum-4));

      if (++ledNum > 31) {
        ledNum = 0;
      }
      delay(delayTimeMS);
    }
  }
}

// Unidirectional pattern of LED with streamers that slows down
void streamer3Pattern() {

  struct RGB CV1, CV2, CV3, CV4;
  const float hueIncrement = 10.0;
  float hueAngle = 360.0;
  int16_t delayTimeMS, ledNum = 0;

  TlcMux_clear();

  delayTimeMS = 20;

  // Do for 32 colors
  for (int c = 0; c < 32; c++) {

    delayTimeMS += 4;

    // Calculate colors at the various percentages of brightness
    createColor(hueAngle, 1.0, &CV1);
    createColor(hueAngle, 0.4, &CV2);
    createColor(hueAngle, 0.2, &CV3);
    createColor(hueAngle, 0.1, &CV4);

    hueAngle -= hueIncrement;

    // For each LED
    for (int i = 0; i < 32; i++) {
      setLEDRGB(adjLEDNum(ledNum),   CV1);
      setLEDRGB(adjLEDNum(ledNum+1), CV2);
      setLEDRGB(adjLEDNum(ledNum+2), CV3);
      setLEDRGB(adjLEDNum(ledNum+3), CV4);
      setLEDOff(adjLEDNum(ledNum+4));

      if (--ledNum < 0) {
        ledNum = 31;
      }
      delay(delayTimeMS);
    }
  }
}

// Bidirectional pattern of LED with streamers
void streamer4Pattern() {

  struct RGB CV1, CV2, CV3, CV4;
  const float hueIncrement = 11.25;
  float hueAngle = 0.0;
  int8_t ledNum;

  boolean direction = false;

  // Do for 32 colors
  for (int c = 0; c < 32; c++) {

    TlcMux_clear();
    delay(500);

    // Toggle direction flag
    direction = !direction;

    // Calculate colors at the various percentages of brightness
    createColor(hueAngle, 1.0, &CV1);
    createColor(hueAngle, 0.4, &CV2);
    createColor(hueAngle, 0.2, &CV3);
    createColor(hueAngle, 0.1, &CV4);

    hueAngle += hueIncrement;

    if (direction) {
      // Forward direction
      ledNum = 3;
    }  
    else   {
      // Reverse direction
      ledNum = 0;
    }

    // For each LED
    for (int i = 0; i < 32; i++) {
      if (direction) {
        // Forward
        setLEDRGB(adjLEDNum(ledNum),   CV1);
        setLEDRGB(adjLEDNum(ledNum-1), CV2);
        setLEDRGB(adjLEDNum(ledNum-2), CV3);
        setLEDRGB(adjLEDNum(ledNum-3), CV4);
        setLEDOff(adjLEDNum(ledNum-4));

        if (++ledNum > 31) {
          ledNum = 0;
        }
      }  
      else  {
        // Reverse
        setLEDRGB(adjLEDNum(ledNum),   CV1);
        setLEDRGB(adjLEDNum(ledNum+1), CV2);
        setLEDRGB(adjLEDNum(ledNum+2), CV3);
        setLEDRGB(adjLEDNum(ledNum+3), CV4);
        setLEDOff(adjLEDNum(ledNum+4));

        if (--ledNum < 0) {
          ledNum = 31;
        }
      }
      delay(80);
    }
  }
}

// Initialization entry point
void setup() {

  // Enable the serial library for debugging
  Serial.begin(9600);

  // Reseed random number generator
  unsigned long time = millis();
  randomSeed(time);

  // Create HSV palette
  createHSVPalette();

  shiftRow = 0;

  // Make all bits of analog port C into digital outputs
  DDRC |= _BV(PC0) | _BV(PC1) | _BV(PC2) | _BV(PC3) | _BV(PC4) | _BV(PC5);

  // Turn off all row enables so LEDs are all off until instructed to turn on
  PORTC = 15;

  // Initialize the TLC5940 mux library
  TlcMux_init();
}

// Main program entry point
void loop() {

  const uint8_t NUMBER_OF_PATTERNS = 24;
  const uint8_t MAX_SPINS = 200;

  uint8_t index, numberOfUsedIndices, spins; 

  // Create a flag for each display pattern
  uint8_t flags[NUMBER_OF_PATTERNS];
  numberOfUsedIndices = NUMBER_OF_PATTERNS;

  TlcMux_clear();

 while (true) {

    // If all patterns have been displayed, start over
    if (numberOfUsedIndices == NUMBER_OF_PATTERNS) {
      // Clear all flags
      memset(flags, 0, NUMBER_OF_PATTERNS);
      numberOfUsedIndices = 0;
    }
    // Set the spin count to zero
    spins = 0;

    // Loop until an unused index is found
    while (true) {  

      // Make sure we don't spin forever just because random
      // number generator won't return a specific index
      if (++spins >= MAX_SPINS) {
        numberOfUsedIndices = NUMBER_OF_PATTERNS;

        // Select my favorite index/pattern
        index = 9;
        break;
      }

      // Pick a candidate pattern index
      index = random(NUMBER_OF_PATTERNS);

      // Check to see if this pattern was previously selected
      if (flags[index] == 0) {
        // This index was not previously used, so mark it so
        flags[index] = 1;

        // Inc number of used indices
        numberOfUsedIndices++;
        break;
      }
    }

    switch(index) {
    case 0:
      redGreenPattern();
      break;
    case 1:
      greenBluePattern();
      break;
    case 2:
      redBluePattern();
      break;
    case 3:
      chase1Pattern(6);
      break;
    case 4:
      chase2Pattern(6);
      break;
    case 5:
      chase3Pattern(6);
      break;
    case 6:
      sequence1Pattern(1);
      break;
    case 7:
      sequence2Pattern(1);
      break;
    case 8:
      sequence3Pattern(4);
      break;
    case 9:
      rotate1Pattern(96);
      break;
    case 10:
      rotate2Pattern(4);
      break;
    case 11:
      rotate3Pattern();
      break;
    case 12:
      convergingLines1Pattern(5);
      break;
    case 13:
      convergingLines2Pattern(5);
      break;
    case 14:
      divergingLines1Pattern(5);
      break;
    case 15:
      divergingLines2Pattern(5);
      break;
    case 16:
      streamer1Pattern();
      break;
    case 17:
      streamer2Pattern();
      break;
    case 18:
      streamer3Pattern();
      break;
    case 19:
      streamer4Pattern();
      break;
    case 20:
      fillInPattern(5);
      break;
    case 21:
      rainPattern(4);
      break;
    case 22:
      flashPattern(10);
      break;
    case 23:
      randomPattern(60);
      break;
    }
  }
}

























